@@ -24,17 +24,18 @@ import (
 	"time"
 
 	"github.com/google/go-cmp/cmp"
+	"go.opencensus.io/resource"
 
 	"k8s.io/apimachinery/pkg/types"
 
 	"knative.dev/pkg/metrics/metricskey"
 	"knative.dev/pkg/metrics/metricstest"
 
+	logtesting "knative.dev/pkg/logging/testing"
 	"knative.dev/serving/pkg/autoscaler/metrics"
 	smetrics "knative.dev/serving/pkg/metrics"
 	"knative.dev/serving/pkg/resources"
 
-	. "knative.dev/pkg/logging/testing"
 	_ "knative.dev/pkg/metrics/testing"
 )
 
@@ -47,12 +48,14 @@ const (
 	testNamespace = "in-this-namespace"
 )
 
-// Standard set of tags for the tests.
-var wantTags = map[string]string{
-	metricskey.LabelConfigurationName: "testConfig",
-	metricskey.LabelNamespaceName:     testNamespace,
-	metricskey.LabelRevisionName:      testRevision,
-	metricskey.LabelServiceName:       "testSvc",
+var wantResource = &resource.Resource{
+	Type: "knative_revision",
+	Labels: map[string]string{
+		metricskey.LabelConfigurationName: "testConfig",
+		metricskey.LabelNamespaceName:     testNamespace,
+		metricskey.LabelRevisionName:      testRevision,
+		metricskey.LabelServiceName:       "testSvc",
+	},
 }
 
 type fakePodCounter struct {
@@ -65,19 +68,103 @@ func (fpc fakePodCounter) ReadyCount() (int, error) {
 	return fpc.readyCount, fpc.err
 }
 
-func TestNewErrorWhenGivenNilReadyPodCounter(t *testing.T) {
-	if _, err := New(testNamespace, testRevision, &metricClient{}, nil,
-		&DeciderSpec{TargetValue: 10}, context.Background()); err == nil {
-		t.Error("Expected error when ReadyPodCounter interface is nil, but got none.")
+func TestAutoscalerScaleDownDelay(t *testing.T) {
+	pc := &fakePodCounter{}
+	metrics := &metricClient{}
+	spec := &DeciderSpec{
+		TargetValue:      10,
+		MaxScaleDownRate: 10,
+		MaxScaleUpRate:   10,
+		PanicThreshold:   100,
+		ScaleDownDelay:   5 * time.Minute,
 	}
+	as := New(context.Background(), testNamespace, testRevision, metrics, pc, spec)
+
+	now := time.Time{}
+
+	t.Run("simple", func(t *testing.T) {
+		// scale up.
+		metrics.SetStableAndPanicConcurrency(40, 40)
+		expectScale(t, as, now.Add(2*time.Second), ScaleResult{
+			ScaleValid:      true,
+			DesiredPodCount: 4,
+			NumActivators:   2,
+		})
+		// five minutes pass at reduced concurrency - should not scale down (less than delay).
+		metrics.SetStableAndPanicConcurrency(0, 0)
+		expectScale(t, as, now.Add(5*time.Minute), ScaleResult{
+			ScaleValid:      true,
+			DesiredPodCount: 4,
+			NumActivators:   2,
+		})
+		// five minutes and 2 seconds pass at reduced concurrency - now we scale down.
+		expectScale(t, as, now.Add(5*time.Minute+2*time.Second), ScaleResult{
+			ScaleValid:      true,
+			DesiredPodCount: 0,
+			NumActivators:   2,
+		})
+	})
+
+	t.Run("gradual", func(t *testing.T) {
+		metrics.SetStableAndPanicConcurrency(40, 40)
+		expectScale(t, as, now.Add(9*time.Minute), ScaleResult{
+			ScaleValid:      true,
+			DesiredPodCount: 4,
+			NumActivators:   2,
+		})
+		metrics.SetStableAndPanicConcurrency(30, 30)
+		expectScale(t, as, now.Add(10*time.Minute), ScaleResult{
+			ScaleValid:      true,
+			DesiredPodCount: 4, // 4 still dominates
+			NumActivators:   2,
+		})
+		metrics.SetStableAndPanicConcurrency(0, 0)
+		expectScale(t, as, now.Add(11*time.Minute), ScaleResult{
+			ScaleValid:      true,
+			DesiredPodCount: 4, // still at 4
+			NumActivators:   2,
+		})
+		expectScale(t, as, now.Add(14*time.Minute), ScaleResult{
+			ScaleValid:      true,
+			DesiredPodCount: 3, // 4 scrolls out, drop to 3
+			NumActivators:   2,
+		})
+		expectScale(t, as, now.Add(15*time.Minute), ScaleResult{
+			ScaleValid:      true,
+			DesiredPodCount: 0, // everything scrolled out, drop to 0
+			NumActivators:   2,
+		})
+	})
 }
 
-func TestNewErrorWhenGivenNilStatsReporter(t *testing.T) {
+func TestAutoscalerScaleDownDelayZero(t *testing.T) {
 	pc := &fakePodCounter{}
-	if _, err := New(testNamespace, testRevision, &metricClient{}, pc,
-		&DeciderSpec{TargetValue: 10}, nil); err == nil {
-		t.Error("Expected error when EndpointsInformer interface is nil, but got none.")
+	metrics := &metricClient{}
+	spec := &DeciderSpec{
+		TargetValue:      10,
+		MaxScaleDownRate: 10,
+		MaxScaleUpRate:   10,
+		PanicThreshold:   100,
+		ScaleDownDelay:   0,
 	}
+	as := New(context.Background(), testNamespace, testRevision, metrics, pc, spec)
+
+	now := time.Time{}
+
+	metrics.SetStableAndPanicConcurrency(40, 40)
+	expectScale(t, as, now, ScaleResult{
+		ScaleValid:      true,
+		DesiredPodCount: 4,
+		NumActivators:   2,
+	})
+	// With zero delay we should immediately scale down (this is not the case
+	// with a 1-element time window delay).
+	metrics.SetStableAndPanicConcurrency(20, 20)
+	expectScale(t, as, now.Add(500*time.Millisecond), ScaleResult{
+		ScaleValid:      true,
+		DesiredPodCount: 2,
+		NumActivators:   2,
+	})
 }
 
 func TestAutoscalerNoDataNoAutoscale(t *testing.T) {
@@ -88,14 +175,14 @@ func TestAutoscalerNoDataNoAutoscale(t *testing.T) {
 		},
 	}
 
-	a := newTestAutoscalerNoPC(t, 10, 100, metrics)
+	a := newTestAutoscalerNoPC(10, 100, metrics)
 	expectScale(t, a, time.Now(), ScaleResult{0, 0, MinActivators, false})
 }
 
 func expectedEBC(totCap, targetBC, recordedConcurrency, numPods float64) int32 {
-	// Extra float64 cast disables fused multiply-subtract to force identical behavior on
-	// all platforms. See floating point section in https://golang.org/ref/spec#Operators.
-	return int32(math.Floor(float64(totCap/targetUtilization*numPods) - targetBC - recordedConcurrency))
+	ebcF := totCap/targetUtilization*numPods - targetBC - recordedConcurrency
+	// We need to floor for negative values.
+	return int32(math.Floor(ebcF))
 }
 
 func expectedNA(a *autoscaler, numP float64) int32 {
@@ -107,52 +194,58 @@ func expectedNA(a *autoscaler, numP float64) int32 {
 func TestAutoscalerStartMetrics(t *testing.T) {
 	defer reset()
 	metrics := &metricClient{StableConcurrency: 50.0, PanicConcurrency: 50.0}
-	newTestAutoscalerWithScalingMetric(t, 10, 100, metrics,
+	newTestAutoscalerWithScalingMetric(10, 100, metrics,
 		"concurrency", true /*startInPanic*/)
-	metricstest.CheckLastValueData(t, panicM.Name(), wantTags, 1)
+	metricstest.AssertMetric(t, metricstest.IntMetric(panicM.Name(), 1, nil).WithResource(wantResource))
 }
 
 func TestAutoscalerMetrics(t *testing.T) {
 	defer reset()
 
 	metrics := &metricClient{StableConcurrency: 50.0, PanicConcurrency: 50.0}
-	a := newTestAutoscalerNoPC(t, 10, 100, metrics)
+	a := newTestAutoscalerNoPC(10, 100, metrics)
 	// Non-panic created autoscaler.
-	metricstest.CheckLastValueData(t, panicM.Name(), wantTags, 0)
+	metricstest.AssertMetric(t, metricstest.IntMetric(panicM.Name(), 0, nil).WithResource(wantResource))
 	ebc := expectedEBC(10, 100, 50, 1)
 	na := expectedNA(a, 1)
 	expectScale(t, a, time.Now(), ScaleResult{5, ebc, na, true})
 	spec := a.currentSpec()
 
-	metricstest.CheckLastValueData(t, stableRequestConcurrencyM.Name(), wantTags, 50)
-	metricstest.CheckLastValueData(t, panicRequestConcurrencyM.Name(), wantTags, 50)
-	metricstest.CheckLastValueData(t, desiredPodCountM.Name(), wantTags, 5)
-	metricstest.CheckLastValueData(t, targetRequestConcurrencyM.Name(), wantTags, spec.TargetValue)
-	metricstest.CheckLastValueData(t, excessBurstCapacityM.Name(), wantTags, float64(ebc))
-	metricstest.CheckLastValueData(t, panicM.Name(), wantTags, 1)
+	wantMetrics := []metricstest.Metric{
+		metricstest.FloatMetric(stableRequestConcurrencyM.Name(), 50, nil).WithResource(wantResource),
+		metricstest.FloatMetric(panicRequestConcurrencyM.Name(), 50, nil).WithResource(wantResource),
+		metricstest.IntMetric(desiredPodCountM.Name(), 5, nil).WithResource(wantResource),
+		metricstest.FloatMetric(targetRequestConcurrencyM.Name(), spec.TargetValue, nil).WithResource(wantResource),
+		metricstest.FloatMetric(excessBurstCapacityM.Name(), float64(ebc), nil).WithResource(wantResource),
+		metricstest.IntMetric(panicM.Name(), 1, nil).WithResource(wantResource),
+	}
+	metricstest.AssertMetric(t, wantMetrics...)
 }
 
 func TestAutoscalerMetricsWithRPS(t *testing.T) {
 	defer reset()
 	metrics := &metricClient{PanicRPS: 99.0, StableRPS: 100}
-	a, _ := newTestAutoscalerWithScalingMetric(t, 10, 100, metrics, "rps", false /*startInPanic*/)
+	a, _ := newTestAutoscalerWithScalingMetric(10, 100, metrics, "rps", false /*startInPanic*/)
 	ebc := expectedEBC(10, 100, 99, 1)
 	na := expectedNA(a, 1)
 	expectScale(t, a, time.Now(), ScaleResult{10, ebc, na, true})
 	spec := a.currentSpec()
 
 	expectScale(t, a, time.Now().Add(61*time.Second), ScaleResult{10, ebc, na, true})
-	metricstest.CheckLastValueData(t, stableRPSM.Name(), wantTags, 100)
-	metricstest.CheckLastValueData(t, panicRPSM.Name(), wantTags, 100)
-	metricstest.CheckLastValueData(t, desiredPodCountM.Name(), wantTags, 10)
-	metricstest.CheckLastValueData(t, targetRPSM.Name(), wantTags, spec.TargetValue)
-	metricstest.CheckLastValueData(t, excessBurstCapacityM.Name(), wantTags, float64(ebc))
-	metricstest.CheckLastValueData(t, panicM.Name(), wantTags, 1)
+	wantMetrics := []metricstest.Metric{
+		metricstest.FloatMetric(stableRPSM.Name(), 100, nil).WithResource(wantResource),
+		metricstest.FloatMetric(panicRPSM.Name(), 100, nil).WithResource(wantResource),
+		metricstest.IntMetric(desiredPodCountM.Name(), 10, nil).WithResource(wantResource),
+		metricstest.FloatMetric(targetRPSM.Name(), spec.TargetValue, nil).WithResource(wantResource),
+		metricstest.FloatMetric(excessBurstCapacityM.Name(), float64(ebc), nil).WithResource(wantResource),
+		metricstest.IntMetric(panicM.Name(), 1, nil).WithResource(wantResource),
+	}
+	metricstest.AssertMetric(t, wantMetrics...)
 }
 
 func TestAutoscalerStableModeIncreaseWithConcurrencyDefault(t *testing.T) {
 	metrics := &metricClient{StableConcurrency: 50.0, PanicConcurrency: 10}
-	a := newTestAutoscalerNoPC(t, 10, 101, metrics)
+	a := newTestAutoscalerNoPC(10, 101, metrics)
 	na := expectedNA(a, 1)
 	expectScale(t, a, time.Now(), ScaleResult{5, expectedEBC(10, 101, 10, 1), na, true})
 
@@ -162,7 +255,7 @@ func TestAutoscalerStableModeIncreaseWithConcurrencyDefault(t *testing.T) {
 
 func TestAutoscalerStableModeIncreaseWithRPS(t *testing.T) {
 	metrics := &metricClient{StableRPS: 50.0, PanicRPS: 50}
-	a, _ := newTestAutoscalerWithScalingMetric(t, 10, 101, metrics, "rps", false /*startInPanic*/)
+	a, _ := newTestAutoscalerWithScalingMetric(10, 101, metrics, "rps", false /*startInPanic*/)
 	na := expectedNA(a, 1)
 	expectScale(t, a, time.Now(), ScaleResult{5, expectedEBC(10, 101, 50, 1), na, true})
 
@@ -174,7 +267,7 @@ func TestAutoscalerStableModeIncreaseWithRPS(t *testing.T) {
 func TestAutoscalerUnpanicAfterSlowIncrease(t *testing.T) {
 	// Do initial jump from 10 to 25 pods.
 	metrics := &metricClient{StableConcurrency: 11, PanicConcurrency: 25}
-	a, pc := newTestAutoscaler(t, 1, 98, metrics)
+	a, pc := newTestAutoscaler(1, 98, metrics)
 	pc.readyCount = 10
 
 	na := expectedNA(a, 10)
@@ -212,7 +305,7 @@ func TestAutoscalerUnpanicAfterSlowIncrease(t *testing.T) {
 func TestAutoscalerExtendPanicWindow(t *testing.T) {
 	// Do initial jump from 10 to 25 pods.
 	metrics := &metricClient{StableConcurrency: 11, PanicConcurrency: 25}
-	a, pc := newTestAutoscaler(t, 1, 98, metrics)
+	a, pc := newTestAutoscaler(1, 98, metrics)
 	pc.readyCount = 10
 
 	na := expectedNA(a, 10)
@@ -237,7 +330,7 @@ func TestAutoscalerExtendPanicWindow(t *testing.T) {
 
 func TestAutoscalerStableModeDecrease(t *testing.T) {
 	metrics := &metricClient{StableConcurrency: 100.0, PanicConcurrency: 100}
-	a, pc := newTestAutoscaler(t, 10, 98, metrics)
+	a, pc := newTestAutoscaler(10, 98, metrics)
 	pc.readyCount = 8
 	na := expectedNA(a, 8)
 	expectScale(t, a, time.Now(), ScaleResult{10, expectedEBC(10, 98, 100, 8), na, true})
@@ -248,7 +341,7 @@ func TestAutoscalerStableModeDecrease(t *testing.T) {
 
 func TestAutoscalerStableModeNoTrafficScaleToZero(t *testing.T) {
 	metrics := &metricClient{StableConcurrency: 1, PanicConcurrency: 0}
-	a := newTestAutoscalerNoPC(t, 10, 75, metrics)
+	a := newTestAutoscalerNoPC(10, 75, metrics)
 	na := expectedNA(a, 1)
 	expectScale(t, a, time.Now(), ScaleResult{1, expectedEBC(10, 75, 0, 1), na, true})
 
@@ -259,9 +352,9 @@ func TestAutoscalerStableModeNoTrafficScaleToZero(t *testing.T) {
 // QPS is increasing exponentially. Each scaling event bring concurrency
 // back to the target level (1.0) but then traffic continues to increase.
 // At 1296 QPS traffic stabilizes.
-func TestAutoscalerPanicModeExponentialTrackAndStablize(t *testing.T) {
+func TestAutoscalerPanicModeExponentialTrackAndStabilize(t *testing.T) {
 	metrics := &metricClient{StableConcurrency: 6, PanicConcurrency: 6}
-	a, pc := newTestAutoscaler(t, 1, 101, metrics)
+	a, pc := newTestAutoscaler(1, 101, metrics)
 	na := expectedNA(a, 1)
 	expectScale(t, a, time.Now(), ScaleResult{6, expectedEBC(1, 101, 6, 1), na, true})
 
@@ -308,37 +401,37 @@ func TestAutoscalerScale(t *testing.T) {
 		wantInvalid bool
 	}{{
 		label:     "AutoscalerNoDataAtZeroNoAutoscale",
-		as:        newTestAutoscalerNoPC(t, 10, 100, &metricClient{}),
+		as:        newTestAutoscalerNoPC(10, 100, &metricClient{}),
 		baseScale: 1,
 		wantScale: 0,
 		wantEBC:   expectedEBC(10, 100, 0, 1),
 	}, {
 		label:     "AutoscalerNoDataAtZeroNoAutoscaleWithExplicitEPs",
-		as:        newTestAutoscalerNoPC(t, 10, 100, &metricClient{}),
+		as:        newTestAutoscalerNoPC(10, 100, &metricClient{}),
 		baseScale: 1,
 		wantScale: 0,
 		wantEBC:   expectedEBC(10, 100, 0, 1),
 	}, {
 		label:     "AutoscalerStableModeUnlimitedTBC",
-		as:        newTestAutoscalerNoPC(t, 181, -1, &metricClient{StableConcurrency: 21.0, PanicConcurrency: 26}),
+		as:        newTestAutoscalerNoPC(181, -1, &metricClient{StableConcurrency: 21.0, PanicConcurrency: 26}),
 		baseScale: 1,
 		wantScale: 1,
 		wantEBC:   -1,
 	}, {
 		label:     "Autoscaler0TBC",
-		as:        newTestAutoscalerNoPC(t, 10, 0, &metricClient{StableConcurrency: 50.0, PanicConcurrency: 49}),
+		as:        newTestAutoscalerNoPC(10, 0, &metricClient{StableConcurrency: 50.0, PanicConcurrency: 49}),
 		baseScale: 1,
 		wantScale: 5,
 		wantEBC:   0,
 	}, {
 		label:     "AutoscalerStableModeNoChange",
-		as:        newTestAutoscalerNoPC(t, 10, 100, &metricClient{StableConcurrency: 50.0, PanicConcurrency: 50}),
+		as:        newTestAutoscalerNoPC(10, 100, &metricClient{StableConcurrency: 50.0, PanicConcurrency: 50}),
 		baseScale: 1,
 		wantScale: 5,
 		wantEBC:   expectedEBC(10, 100, 50, 1),
 	}, {
 		label: "AutoscalerPanicStableLargerThanPanic",
-		as:    newTestAutoscalerNoPC(t, 1, 100, &metricClient{StableConcurrency: 50, PanicConcurrency: 30}),
+		as:    newTestAutoscalerNoPC(1, 100, &metricClient{StableConcurrency: 50, PanicConcurrency: 30}),
 		prepFunc: func(a *autoscaler) {
 			a.panicTime = time.Now().Add(-5 * time.Second)
 			a.maxPanicPods = 5
@@ -348,7 +441,7 @@ func TestAutoscalerScale(t *testing.T) {
 		wantEBC:   expectedEBC(1, 100, 30, 5),
 	}, {
 		label: "AutoscalerPanicStableLessThanPanic",
-		as:    newTestAutoscalerNoPC(t, 1, 100, &metricClient{StableConcurrency: 20, PanicConcurrency: 30}),
+		as:    newTestAutoscalerNoPC(1, 100, &metricClient{StableConcurrency: 20, PanicConcurrency: 30}),
 		prepFunc: func(a *autoscaler) {
 			a.panicTime = time.Now().Add(-5 * time.Second)
 			a.maxPanicPods = 5
@@ -358,13 +451,13 @@ func TestAutoscalerScale(t *testing.T) {
 		wantEBC:   expectedEBC(1, 100, 30, 5),
 	}, {
 		label:     "AutoscalerStableModeNoChangeAlreadyScaled",
-		as:        newTestAutoscalerNoPC(t, 10, 100, &metricClient{StableConcurrency: 50.0, PanicConcurrency: 50}),
+		as:        newTestAutoscalerNoPC(10, 100, &metricClient{StableConcurrency: 50.0, PanicConcurrency: 50}),
 		baseScale: 5,
 		wantScale: 5,
 		wantEBC:   expectedEBC(10, 100, 50, 5),
 	}, {
 		label: "AutoscalerStableModeIncreaseWithSmallScaleUpRate",
-		as: newTestAutoscalerNoPC(t, 1 /* target */, 1982 /* TBC */, &metricClient{
+		as: newTestAutoscalerNoPC(1 /* target */, 1982 /* TBC */, &metricClient{
 			StableConcurrency: 3,
 			PanicConcurrency:  3.1,
 		}),
@@ -376,7 +469,7 @@ func TestAutoscalerScale(t *testing.T) {
 		wantEBC:   expectedEBC(1, 1982, 3.1, 2),
 	}, {
 		label:     "AutoscalerStableModeDecreaseWithSmallScaleDownRate",
-		as:        newTestAutoscalerNoPC(t, 10 /* target */, 1982 /* TBC */, &metricClient{StableConcurrency: 1, PanicConcurrency: 1}),
+		as:        newTestAutoscalerNoPC(10 /* target */, 1982 /* TBC */, &metricClient{StableConcurrency: 1, PanicConcurrency: 1}),
 		baseScale: 100,
 		prepFunc: func(a *autoscaler) {
 			a.deciderSpec.MaxScaleDownRate = 1.1
@@ -384,8 +477,8 @@ func TestAutoscalerScale(t *testing.T) {
 		wantScale: 90,
 		wantEBC:   expectedEBC(10, 1982, 1, 100),
 	}, {
-		label:     "AutoscalerStableModeDecreseNonReachable",
-		as:        newTestAutoscalerNoPC(t, 10 /* target */, 1982 /* TBC */, &metricClient{StableConcurrency: 1, PanicConcurrency: 1}),
+		label:     "AutoscalerStableModeDecreaseNonReachable",
+		as:        newTestAutoscalerNoPC(10 /* target */, 1982 /* TBC */, &metricClient{StableConcurrency: 1, PanicConcurrency: 1}),
 		baseScale: 100,
 		prepFunc: func(a *autoscaler) {
 			a.deciderSpec.MaxScaleDownRate = 1.1
@@ -395,7 +488,7 @@ func TestAutoscalerScale(t *testing.T) {
 		wantEBC:   expectedEBC(10, 1982, 1, 100),
 	}, {
 		label:     "AutoscalerPanicModeDoublePodCount",
-		as:        newTestAutoscalerNoPC(t, 10, 84, &metricClient{StableConcurrency: 50, PanicConcurrency: 100}),
+		as:        newTestAutoscalerNoPC(10, 84, &metricClient{StableConcurrency: 50, PanicConcurrency: 100}),
 		baseScale: 1,
 		// PanicConcurrency takes precedence.
 		wantScale: 10,
@@ -416,7 +509,7 @@ func TestAutoscalerScale(t *testing.T) {
 
 func TestAutoscalerPanicThenUnPanicScaleDown(t *testing.T) {
 	metrics := &metricClient{StableConcurrency: 100, PanicConcurrency: 100}
-	a, pc := newTestAutoscaler(t, 10, 93, metrics)
+	a, pc := newTestAutoscaler(10, 93, metrics)
 	na := expectedNA(a, 1)
 	expectScale(t, a, time.Now(), ScaleResult{10, expectedEBC(10, 93, 100, 1), na, true})
 	pc.readyCount = 10
@@ -436,7 +529,7 @@ func TestAutoscalerPanicThenUnPanicScaleDown(t *testing.T) {
 
 func TestAutoscalerRateLimitScaleUp(t *testing.T) {
 	metrics := &metricClient{StableConcurrency: 1000, PanicConcurrency: 1001}
-	a, pc := newTestAutoscaler(t, 10, 61, metrics)
+	a, pc := newTestAutoscaler(10, 61, metrics)
 	na := expectedNA(a, 1)
 
 	// Need 100 pods but only scale x10
@@ -450,7 +543,7 @@ func TestAutoscalerRateLimitScaleUp(t *testing.T) {
 
 func TestAutoscalerRateLimitScaleDown(t *testing.T) {
 	metrics := &metricClient{StableConcurrency: 1, PanicConcurrency: 1}
-	a, pc := newTestAutoscaler(t, 10, 61, metrics)
+	a, pc := newTestAutoscaler(10, 61, metrics)
 
 	// Need 1 pods but can only scale down ten times, to 10.
 	pc.readyCount = 100
@@ -465,16 +558,16 @@ func TestAutoscalerRateLimitScaleDown(t *testing.T) {
 
 func TestCantCountPods(t *testing.T) {
 	metrics := &metricClient{StableConcurrency: 1000, PanicConcurrency: 888}
-	a, pc := newTestAutoscaler(t, 10, 81, metrics)
+	a, pc := newTestAutoscaler(10, 81, metrics)
 	pc.err = errors.New("peaches-in-regalia")
-	if got, want := a.Scale(context.Background(), time.Now()), invalidSR; !cmp.Equal(got, want) {
+	if got, want := a.Scale(logtesting.TestLogger(t), time.Now()), invalidSR; !cmp.Equal(got, want) {
 		t.Errorf("Scale = %v, want: %v", got, want)
 	}
 }
 
 func TestAutoscalerUseOnePodAsMinimumIfEndpointsNotFound(t *testing.T) {
 	metrics := &metricClient{StableConcurrency: 1000, PanicConcurrency: 888}
-	a, pc := newTestAutoscaler(t, 10, 81, metrics)
+	a, pc := newTestAutoscaler(10, 81, metrics)
 
 	pc.readyCount = 0
 	// 2*10 as the rate limited if we can get the actual pods number.
@@ -484,7 +577,7 @@ func TestAutoscalerUseOnePodAsMinimumIfEndpointsNotFound(t *testing.T) {
 
 func TestAutoscalerUpdateTarget(t *testing.T) {
 	metrics := &metricClient{StableConcurrency: 100, PanicConcurrency: 101}
-	a, pc := newTestAutoscaler(t, 10, 77, metrics)
+	a, pc := newTestAutoscaler(10, 77, metrics)
 	na := expectedNA(a, 1)
 	expectScale(t, a, time.Now(), ScaleResult{10, expectedEBC(10, 77, 101, 1), na, true})
 
@@ -504,20 +597,19 @@ func TestAutoscalerUpdateTarget(t *testing.T) {
 }
 
 // For table tests and tests that don't care about changing scale.
-func newTestAutoscalerNoPC(t *testing.T, targetValue, targetBurstCapacity float64,
+func newTestAutoscalerNoPC(targetValue, targetBurstCapacity float64,
 	metrics metrics.MetricClient) *autoscaler {
-	a, _ := newTestAutoscaler(t, targetValue, targetBurstCapacity, metrics)
+	a, _ := newTestAutoscaler(targetValue, targetBurstCapacity, metrics)
 	return a
 }
 
-func newTestAutoscaler(t *testing.T, targetValue, targetBurstCapacity float64,
+func newTestAutoscaler(targetValue, targetBurstCapacity float64,
 	metrics metrics.MetricClient) (*autoscaler, *fakePodCounter) {
-	return newTestAutoscalerWithScalingMetric(t, targetValue, targetBurstCapacity,
+	return newTestAutoscalerWithScalingMetric(targetValue, targetBurstCapacity,
 		metrics, "concurrency", false /*panic*/)
 }
 
-func newTestAutoscalerWithScalingMetric(t *testing.T, targetValue, targetBurstCapacity float64, metrics metrics.MetricClient, metric string, startInPanic bool) (*autoscaler, *fakePodCounter) {
-	t.Helper()
+func newTestAutoscalerWithScalingMetric(targetValue, targetBurstCapacity float64, metrics metrics.MetricClient, metric string, startInPanic bool) (*autoscaler, *fakePodCounter) {
 	deciderSpec := &DeciderSpec{
 		ScalingMetric:       metric,
 		TargetValue:         targetValue,
@@ -536,18 +628,26 @@ func newTestAutoscalerWithScalingMetric(t *testing.T, targetValue, targetBurstCa
 	if startInPanic {
 		pc.readyCount = 2
 	}
-	ctx, err := smetrics.RevisionContext(testNamespace, "testSvc", "testConfig", testRevision)
-	if err != nil {
-		t.Fatal("Error creating context:", err)
-	}
-	return newAutoscaler(testNamespace, testRevision,
-		metrics, pc, deciderSpec, ctx), pc
+	ctx := smetrics.RevisionContext(testNamespace, "testSvc", "testConfig", testRevision)
+	return newAutoscaler(ctx, testNamespace, testRevision, metrics, pc, deciderSpec, nil), pc
+}
+
+// approxEquateInt32 equates int32s with given path with ±-1 tolerance.
+// This is needed due to rounding errors across various platforms.
+func approxEquateInt32(field string) cmp.Option {
+	eqOpt := cmp.Comparer(func(a, b int32) bool {
+		d := a - b
+		return d <= 1 && d >= -1
+	})
+	return cmp.FilterPath(func(p cmp.Path) bool {
+		return p.String() == field
+	}, eqOpt)
 }
 
 func expectScale(t *testing.T, a UniScaler, now time.Time, want ScaleResult) {
 	t.Helper()
-	got := a.Scale(TestContextWithLogger(t), now)
-	if !cmp.Equal(got, want) {
+	got := a.Scale(logtesting.TestLogger(t), now)
+	if !cmp.Equal(got, want, approxEquateInt32("ExcessBurstCapacity")) {
 		t.Error("ScaleResult mismatch(-want,+got):\n", cmp.Diff(want, got))
 	}
 }
@@ -567,7 +667,7 @@ func TestStartInPanicMode(t *testing.T) {
 	pc := &fakePodCounter{}
 	for i := 0; i < 2; i++ {
 		pc.readyCount = i
-		a := newAutoscaler(testNamespace, testRevision, metrics, pc, deciderSpec, context.Background())
+		a := newAutoscaler(context.Background(), testNamespace, testRevision, metrics, pc, deciderSpec, nil)
 		if !a.panicTime.IsZero() {
 			t.Errorf("Create at scale %d had panic mode on", i)
 		}
@@ -578,7 +678,7 @@ func TestStartInPanicMode(t *testing.T) {
 
 	// Now start with 2 and make sure we're in panic mode.
 	pc.readyCount = 2
-	a := newAutoscaler(testNamespace, testRevision, metrics, pc, deciderSpec, context.Background())
+	a := newAutoscaler(context.Background(), testNamespace, testRevision, metrics, pc, deciderSpec, nil)
 	if a.panicTime.IsZero() {
 		t.Error("Create at scale 2 had panic mode off")
 	}
@@ -600,7 +700,7 @@ func TestNewFail(t *testing.T) {
 	}
 
 	pc := fakePodCounter{err: errors.New("starlight")}
-	a := newAutoscaler(testNamespace, testRevision, metrics, pc, deciderSpec, context.Background())
+	a := newAutoscaler(context.Background(), testNamespace, testRevision, metrics, pc, deciderSpec, nil)
 	if got, want := int(a.maxPanicPods), 0; got != want {
 		t.Errorf("maxPanicPods = %d, want: 0", got)
 	}
@@ -657,3 +757,13 @@ func (mc *metricClient) StableAndPanicRPS(key types.NamespacedName, now time.Tim
 	}
 	return mc.StableRPS, mc.PanicRPS, err
 }
+
+func BenchmarkAutoscaler(b *testing.B) {
+	metrics := &metricClient{StableConcurrency: 50.0, PanicConcurrency: 10}
+	a := newTestAutoscalerNoPC(10, 101, metrics)
+	now := time.Now()
+
+	for i := 0; i < b.N; i++ {
+		a.Scale(logtesting.TestLogger(b), now)
+	}
+}
