@@ -5,7 +5,7 @@ Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
 
-    https://www.apache.org/licenses/LICENSE-2.0
+    http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
@@ -17,21 +17,22 @@ limitations under the License.
 package serverlessservice
 
 import (
+	"context"
 	"testing"
 	"time"
 
 	"github.com/google/go-cmp/cmp"
 	"golang.org/x/sync/errgroup"
 
+	networkingv1alpha1 "knative.dev/networking/pkg/client/clientset/versioned/typed/networking/v1alpha1"
 	fakenetworkingclient "knative.dev/networking/pkg/client/injection/client/fake"
 	fakekubeclient "knative.dev/pkg/client/injection/kube/client/fake"
 	fakeendpointsinformer "knative.dev/pkg/client/injection/kube/informers/core/v1/endpoints/fake"
 	"knative.dev/pkg/configmap"
-	"knative.dev/pkg/controller"
 	fakedynamicclient "knative.dev/pkg/injection/clients/dynamicclient/fake"
 	"knative.dev/serving/pkg/client/injection/ducks/autoscaling/v1alpha1/podscalable"
 
-	"k8s.io/apimachinery/pkg/labels"
+	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/util/wait"
 
@@ -60,32 +61,31 @@ func TestGlobalResyncOnActivatorChange(t *testing.T) {
 	grp := errgroup.Group{}
 
 	kubeClnt := fakekubeclient.Get(ctx)
+	epsInformer := fakeendpointsinformer.Get(ctx)
 
 	// Create activator endpoints.
 	aEps := activatorEndpoints(WithSubsets)
-	if _, err := kubeClnt.CoreV1().Endpoints(aEps.Namespace).Create(aEps); err != nil {
-		t.Fatal("Error creating activator endpoints:", err)
-	}
+	kubeClnt.CoreV1().Endpoints(aEps.Namespace).Create(ctx, aEps, metav1.CreateOptions{})
+	epsInformer.Informer().GetIndexer().Add(aEps)
 
 	// Private endpoints are supposed to exist, since we're using selector mode for the service.
 	privEps := endpointspriv(ns1, sks1)
-	if _, err := kubeClnt.CoreV1().Endpoints(privEps.Namespace).Create(privEps); err != nil {
-		t.Fatal("Error creating private endpoints:", err)
-	}
+	kubeClnt.CoreV1().Endpoints(privEps.Namespace).Create(ctx, privEps, metav1.CreateOptions{})
+	epsInformer.Informer().GetIndexer().Add(privEps)
+
 	// This is passive, so no endpoints.
 	privEps = endpointspriv(ns2, sks2, withOtherSubsets)
-	if _, err := kubeClnt.CoreV1().Endpoints(privEps.Namespace).Create(privEps); err != nil {
-		t.Fatal("Error creating private endpoints:", err)
-	}
+	kubeClnt.CoreV1().Endpoints(privEps.Namespace).Create(ctx, privEps, metav1.CreateOptions{})
+	epsInformer.Informer().GetIndexer().Add(privEps)
 
-	waitInformers, err := controller.RunInformers(ctx.Done(), informers...)
+	waitInformers, err := RunAndSyncInformers(ctx, informers...)
 	if err != nil {
 		t.Fatal("Error starting informers:", err)
 	}
 	defer func() {
 		cancel()
 		if err := grp.Wait(); err != nil {
-			t.Fatal("Error waiting for contoller to terminate:", err)
+			t.Fatal("Error waiting for controller to terminate:", err)
 		}
 		waitInformers()
 	}()
@@ -94,36 +94,36 @@ func TestGlobalResyncOnActivatorChange(t *testing.T) {
 		return ctrl.Run(1, ctx.Done())
 	})
 
+	networking := fakenetworkingclient.Get(ctx).NetworkingV1alpha1()
+
 	// Inactive, will reconcile.
 	sksObj1 := SKS(ns1, sks1, WithPrivateService, WithPubService, WithDeployRef(sks1), WithProxyMode)
+	sksObj1.Generation = 1
 	// Active, should not visibly reconcile.
 	sksObj2 := SKS(ns2, sks2, WithPrivateService, WithPubService, WithDeployRef(sks2), markHappy)
+	sksObj2.Generation = 1
 
-	if _, err := fakenetworkingclient.Get(ctx).NetworkingV1alpha1().ServerlessServices(ns1).Create(sksObj1); err != nil {
+	if _, err := networking.ServerlessServices(ns1).Create(ctx, sksObj1, metav1.CreateOptions{}); err != nil {
 		t.Fatal("Error creating SKS1:", err)
 	}
-	if _, err := fakenetworkingclient.Get(ctx).NetworkingV1alpha1().ServerlessServices(ns2).Create(sksObj2); err != nil {
+	if _, err := networking.ServerlessServices(ns2).Create(ctx, sksObj2, metav1.CreateOptions{}); err != nil {
 		t.Fatal("Error creating SKS2:", err)
 	}
 
-	eps := fakeendpointsinformer.Get(ctx).Lister()
-	if err := wait.PollImmediate(10*time.Millisecond, 5*time.Second, func() (bool, error) {
-		l, err := eps.List(labels.Everything())
-		return len(l) >= 4, err
-	}); err != nil {
-		t.Fatal("Failed to see endpoint creation:", err)
-	}
-	t.Log("Updating the activator endpoints now...")
+	// Wait for the SKSs to be reconciled
+	waitForObservedGen(ctx, networking, ns1, sks1, 1)
+	waitForObservedGen(ctx, networking, ns2, sks2, 1)
 
+	t.Log("Updating the activator endpoints now...")
 	// Now that we have established the baseline, update the activator endpoints.
 	aEps = activatorEndpoints(withOtherSubsets)
-	if _, err := kubeClnt.CoreV1().Endpoints(aEps.Namespace).Update(aEps); err != nil {
+	if _, err := kubeClnt.CoreV1().Endpoints(aEps.Namespace).Update(ctx, aEps, metav1.UpdateOptions{}); err != nil {
 		t.Fatal("Error creating activator endpoints:", err)
 	}
 
 	// Actively wait for the endpoints to change their value.
-	if err := wait.PollImmediate(10*time.Millisecond, 3*time.Second, func() (bool, error) {
-		ep, err := eps.Endpoints(ns1).Get(sks1)
+	if err := wait.PollImmediate(25*time.Millisecond, 5*time.Second, func() (bool, error) {
+		ep, err := epsInformer.Lister().Endpoints(ns1).Get(sks1)
 		if err != nil {
 			return false, err
 		}
@@ -135,3 +135,13 @@ func TestGlobalResyncOnActivatorChange(t *testing.T) {
 		t.Fatal("Failed to see Public Endpoints propagation:", err)
 	}
 }
+
+func waitForObservedGen(ctx context.Context, client networkingv1alpha1.NetworkingV1alpha1Interface, ns, name string, generation int64) error {
+	return wait.PollImmediate(10*time.Millisecond, 5*time.Second, func() (bool, error) {
+		sks, err := client.ServerlessServices(ns).Get(ctx, name, metav1.GetOptions{})
+		if err != nil {
+			return false, err
+		}
+		return sks.Status.ObservedGeneration == 1, nil
+	})
+}
