@@ -19,28 +19,31 @@ limitations under the License.
 package v1
 
 import (
+	"context"
 	"testing"
 
 	"github.com/google/go-cmp/cmp"
 	"github.com/google/go-cmp/cmp/cmpopts"
+
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"knative.dev/pkg/ptr"
-	pkgTest "knative.dev/pkg/test"
+	pkgtest "knative.dev/pkg/test"
 	v1 "knative.dev/serving/pkg/apis/serving/v1"
 	"knative.dev/serving/test"
 	v1test "knative.dev/serving/test/v1"
 
+	corev1 "k8s.io/api/core/v1"
 	rtesting "knative.dev/serving/pkg/testing/v1"
 )
 
-// TestService tests both Creation and Update paths for a service. The test performs a series of Update/Validate steps to ensure that
+// TestServiceCreateAndUpdate tests both Creation and Update paths for a service. The test performs a series of Update/Validate steps to ensure that
 // the service transitions as expected during each step.
 // Currently the test performs the following updates:
 // 1. Update Container Image
 // 2. Update Metadata
 //    a. Update Labels
 //    b. Update Annotations
-func TestService(t *testing.T) {
+func TestServiceCreateAndUpdate(t *testing.T) {
 	t.Parallel()
 	clients := test.Setup(t)
 
@@ -73,17 +76,17 @@ func TestService(t *testing.T) {
 	}
 
 	if err := validateAnnotations(objects); err != nil {
-		t.Errorf("Service annotations are incorrect: %v", err)
+		t.Error("Service annotations are incorrect:", err)
 	}
 
 	// We start a background prober to test if Route is always healthy even during Route update.
-	prober := test.RunRouteProber(t.Logf, clients, names.URL, test.AddRootCAtoTransport(t.Logf, clients, test.ServingFlags.Https))
+	prober := test.RunRouteProber(t.Logf, clients, names.URL, test.AddRootCAtoTransport(context.Background(), t.Logf, clients, test.ServingFlags.HTTPS))
 	defer test.AssertProberDefault(t, prober)
 
 	// Update Container Image
 	t.Log("Updating the Service to use a different image.")
 	names.Image = test.PizzaPlanet2
-	image2 := pkgTest.ImagePath(names.Image)
+	image2 := pkgtest.ImagePath(names.Image)
 	if _, err := v1test.PatchService(t, clients, objects.Service, rtesting.WithServiceImage(image2)); err != nil {
 		t.Fatalf("Patch update for Service %s with new image %s failed: %v", names.Service, image2, err)
 	}
@@ -125,7 +128,7 @@ func TestService(t *testing.T) {
 	}
 
 	// Update Metadata (Annotations)
-	t.Logf("Updating annotations of RevisionTemplateSpec for service %s", names.Service)
+	t.Log("Updating annotations of RevisionTemplateSpec for service", names.Service)
 	metadata = metav1.ObjectMeta{
 		Annotations: map[string]string{
 			"annotationA": "123",
@@ -219,17 +222,17 @@ func TestServiceBYOName(t *testing.T) {
 	}
 
 	if err := validateAnnotations(objects); err != nil {
-		t.Errorf("Service annotations are incorrect: %v", err)
+		t.Error("Service annotations are incorrect:", err)
 	}
 
 	// We start a background prober to test if Route is always healthy even during Route update.
-	prober := test.RunRouteProber(t.Logf, clients, names.URL, test.AddRootCAtoTransport(t.Logf, clients, test.ServingFlags.Https))
+	prober := test.RunRouteProber(t.Logf, clients, names.URL, test.AddRootCAtoTransport(context.Background(), t.Logf, clients, test.ServingFlags.HTTPS))
 	defer test.AssertProberDefault(t, prober)
 
 	// Update Container Image
 	t.Log("Updating the Service to use a different image.")
 	names.Image = test.PizzaPlanet2
-	image2 := pkgTest.ImagePath(names.Image)
+	image2 := pkgtest.ImagePath(names.Image)
 	if _, err := v1test.PatchService(t, clients, objects.Service, rtesting.WithServiceImage(image2)); err == nil {
 		t.Fatalf("Patch update for Service %s didn't fail.", names.Service)
 	}
@@ -246,8 +249,8 @@ func TestServiceWithTrafficSplit(t *testing.T) {
 	t.Parallel()
 	// Create Initial Service
 	clients := test.Setup(t)
-	releaseImagePath2 := pkgTest.ImagePath(test.PizzaPlanet2)
-	releaseImagePath3 := pkgTest.ImagePath(test.HelloWorld)
+	releaseImagePath2 := pkgtest.ImagePath(test.PizzaPlanet2)
+	releaseImagePath3 := pkgtest.ImagePath(test.HelloWorld)
 	names := test.ResourceNames{
 		Service: test.ObjectNameForTest(t),
 		Image:   test.PizzaPlanet1,
@@ -272,7 +275,7 @@ func TestServiceWithTrafficSplit(t *testing.T) {
 		t.Fatal("Release shape is incorrect:", err)
 	}
 	if err := validateAnnotations(objects); err != nil {
-		t.Errorf("Service annotations are incorrect: %v", err)
+		t.Error("Service annotations are incorrect:", err)
 	}
 	firstRevision := names.Revision
 
@@ -303,6 +306,7 @@ func TestServiceWithTrafficSplit(t *testing.T) {
 			Tag:            "latest",
 			RevisionName:   objects.Config.Status.LatestReadyRevisionName,
 			LatestRevision: ptr.Bool(true),
+			Percent:        ptr.Int64(0),
 		},
 	}
 	t.Log("Waiting for Service to become ready with the new shape.")
@@ -312,10 +316,15 @@ func TestServiceWithTrafficSplit(t *testing.T) {
 
 	t.Log("Service traffic should go to the first revision and be available on two names traffic targets: 'current' and 'latest'")
 	if err := validateDomains(t, clients,
-		names.URL,
+		names.Service,
 		[]string{expectedFirstRev},
-		[]string{"latest", "current"},
-		[]string{expectedFirstRev, expectedFirstRev}); err != nil {
+		[]tagExpectation{{
+			tag:              "latest",
+			expectedResponse: expectedFirstRev,
+		}, {
+			tag:              "current",
+			expectedResponse: expectedFirstRev,
+		}}); err != nil {
 		t.Fatal(err)
 	}
 
@@ -336,6 +345,7 @@ func TestServiceWithTrafficSplit(t *testing.T) {
 		Tag:            "latest",
 		RevisionName:   secondRevision,
 		LatestRevision: ptr.Bool(true),
+		Percent:        ptr.Int64(0),
 	}
 	t.Log("Waiting for Service to become ready with the new shape.")
 	if err := waitForDesiredTrafficShape(t, names.Service, desiredTrafficShape, clients); err != nil {
@@ -344,10 +354,15 @@ func TestServiceWithTrafficSplit(t *testing.T) {
 
 	t.Log("Since the Service is using release the Route will not be updated, but new revision will be available at 'latest'")
 	if err := validateDomains(t, clients,
-		names.URL,
+		names.Service,
 		[]string{expectedFirstRev},
-		[]string{"latest", "current"},
-		[]string{expectedSecondRev, expectedFirstRev}); err != nil {
+		[]tagExpectation{{
+			tag:              "latest",
+			expectedResponse: expectedSecondRev,
+		}, {
+			tag:              "current",
+			expectedResponse: expectedFirstRev,
+		}}); err != nil {
 		t.Fatal(err)
 	}
 
@@ -363,8 +378,7 @@ func TestServiceWithTrafficSplit(t *testing.T) {
 			RevisionName: secondRevision,
 			Percent:      ptr.Int64(50),
 		}, {
-			Tag:     "latest",
-			Percent: nil,
+			Tag: "latest",
 		}},
 	})
 	if err != nil {
@@ -388,6 +402,7 @@ func TestServiceWithTrafficSplit(t *testing.T) {
 			Tag:            "latest",
 			RevisionName:   secondRevision,
 			LatestRevision: ptr.Bool(true),
+			Percent:        ptr.Int64(0),
 		},
 	}
 	t.Log("Waiting for Service to become ready with the new shape.")
@@ -397,10 +412,18 @@ func TestServiceWithTrafficSplit(t *testing.T) {
 
 	t.Log("Traffic should be split between the two revisions and available on three named traffic targets, 'current', 'candidate', and 'latest'")
 	if err := validateDomains(t, clients,
-		names.URL,
+		names.Service,
 		[]string{expectedFirstRev, expectedSecondRev},
-		[]string{"candidate", "latest", "current"},
-		[]string{expectedSecondRev, expectedSecondRev, expectedFirstRev}); err != nil {
+		[]tagExpectation{{
+			tag:              "candidate",
+			expectedResponse: expectedSecondRev,
+		}, {
+			tag:              "latest",
+			expectedResponse: expectedSecondRev,
+		}, {
+			tag:              "current",
+			expectedResponse: expectedFirstRev,
+		}}); err != nil {
 		t.Fatal(err)
 	}
 
@@ -419,6 +442,7 @@ func TestServiceWithTrafficSplit(t *testing.T) {
 		Tag:            "latest",
 		RevisionName:   thirdRevision,
 		LatestRevision: ptr.Bool(true),
+		Percent:        ptr.Int64(0),
 	}
 	t.Log("Waiting for Service to become ready with the new shape.")
 	if err := waitForDesiredTrafficShape(t, names.Service, desiredTrafficShape, clients); err != nil {
@@ -427,10 +451,18 @@ func TestServiceWithTrafficSplit(t *testing.T) {
 
 	t.Log("Traffic should remain between the two images, and the new revision should be available on the named traffic target 'latest'")
 	if err := validateDomains(t, clients,
-		names.URL,
+		names.Service,
 		[]string{expectedFirstRev, expectedSecondRev},
-		[]string{"latest", "candidate", "current"},
-		[]string{expectedThirdRev, expectedSecondRev, expectedFirstRev}); err != nil {
+		[]tagExpectation{{
+			tag:              "latest",
+			expectedResponse: expectedThirdRev,
+		}, {
+			tag:              "candidate",
+			expectedResponse: expectedSecondRev,
+		}, {
+			tag:              "current",
+			expectedResponse: expectedFirstRev,
+		}}); err != nil {
 		t.Fatal(err)
 	}
 
@@ -456,7 +488,7 @@ func TestServiceWithTrafficSplit(t *testing.T) {
 
 	// Verify in the end it's still the case.
 	if err := validateAnnotations(objects); err != nil {
-		t.Errorf("Service annotations are incorrect: %v", err)
+		t.Error("Service annotations are incorrect:", err)
 	}
 
 	// `candidate` now points to the latest.
@@ -472,10 +504,18 @@ func TestServiceWithTrafficSplit(t *testing.T) {
 	}
 
 	if err := validateDomains(t, clients,
-		names.URL,
+		names.Service,
 		[]string{expectedFirstRev, expectedThirdRev},
-		[]string{"latest", "candidate", "current"},
-		[]string{expectedThirdRev, expectedThirdRev, expectedFirstRev}); err != nil {
+		[]tagExpectation{{
+			tag:              "latest",
+			expectedResponse: expectedThirdRev,
+		}, {
+			tag:              "candidate",
+			expectedResponse: expectedThirdRev,
+		}, {
+			tag:              "current",
+			expectedResponse: expectedFirstRev,
+		}}); err != nil {
 		t.Fatal(err)
 	}
 }
@@ -505,7 +545,7 @@ func TestAnnotationPropagation(t *testing.T) {
 	}
 
 	if err := validateAnnotations(objects); err != nil {
-		t.Errorf("Annotations are incorrect: %v", err)
+		t.Error("Annotations are incorrect:", err)
 	}
 
 	if objects.Service, err = v1test.PatchService(t, clients, objects.Service,
@@ -516,7 +556,7 @@ func TestAnnotationPropagation(t *testing.T) {
 	// Updating metadata does not trigger revision or generation
 	// change, so let's generate a change that we can watch.
 	t.Log("Updating the Service to use a different image.")
-	image2 := pkgTest.ImagePath(test.PizzaPlanet2)
+	image2 := pkgtest.ImagePath(test.PizzaPlanet2)
 	if _, err := v1test.PatchService(t, clients, objects.Service, rtesting.WithServiceImage(image2)); err != nil {
 		t.Fatalf("Patch update for Service %s with new image %s failed: %v", names.Service, image2, err)
 	}
@@ -533,12 +573,12 @@ func TestAnnotationPropagation(t *testing.T) {
 	}
 	objects, err = v1test.GetResourceObjects(clients, names)
 	if err != nil {
-		t.Errorf("Error getting objects: %v", err)
+		t.Error("Error getting objects:", err)
 	}
 
 	// Now we can validate the annotations.
 	if err := validateAnnotations(objects, "juicy"); err != nil {
-		t.Errorf("Annotations are incorrect: %v", err)
+		t.Error("Annotations are incorrect:", err)
 	}
 
 	if objects.Service, err = v1test.PatchService(t, clients, objects.Service,
@@ -549,7 +589,7 @@ func TestAnnotationPropagation(t *testing.T) {
 	// Updating metadata does not trigger revision or generation
 	// change, so let's generate a change that we can watch.
 	t.Log("Updating the Service to use a different image.")
-	image3 := pkgTest.ImagePath(test.HelloWorld)
+	image3 := pkgtest.ImagePath(test.HelloWorld)
 	if _, err := v1test.PatchService(t, clients, objects.Service, rtesting.WithServiceImage(image3)); err != nil {
 		t.Fatalf("Patch update for Service %s with new image %s failed: %v", names.Service, image3, err)
 	}
@@ -566,12 +606,12 @@ func TestAnnotationPropagation(t *testing.T) {
 	}
 	objects, err = v1test.GetResourceObjects(clients, names)
 	if err != nil {
-		t.Errorf("Error getting objects: %v", err)
+		t.Error("Error getting objects:", err)
 	}
 
 	// Now we can validate the annotations.
 	if err := validateAnnotations(objects); err != nil {
-		t.Errorf("Annotations are incorrect: %v", err)
+		t.Error("Annotations are incorrect:", err)
 	}
 	if _, ok := objects.Config.Annotations["juicy"]; ok {
 		t.Error("Config still has `juicy` annotation")
@@ -580,3 +620,48 @@ func TestAnnotationPropagation(t *testing.T) {
 		t.Error("Route still has `juicy` annotation")
 	}
 }
+
+// TestServiceCreateWithMultipleContainers tests both Creation paths for a service.
+// The test performs a series of Validate steps to ensure that the service transitions as expected during each step.
+func TestServiceCreateWithMultipleContainers(t *testing.T) {
+	if !test.ServingFlags.EnableBetaFeatures {
+		t.Skip()
+	}
+	t.Parallel()
+	clients := test.Setup(t)
+
+	names := test.ResourceNames{
+		Service: test.ObjectNameForTest(t),
+		Image:   test.ServingContainer,
+		Sidecars: []string{
+			test.SidecarContainer,
+		},
+	}
+
+	// Clean up on test failure or interrupt
+	test.EnsureTearDown(t, clients, &names)
+	containers := []corev1.Container{{
+		Image: pkgtest.ImagePath(names.Image),
+		Ports: []corev1.ContainerPort{{
+			ContainerPort: 8881,
+		}},
+	}, {
+		Image: pkgtest.ImagePath(names.Sidecars[0]),
+	}}
+
+	// Setup initial Service
+	if _, err := v1test.CreateServiceReady(t, clients, &names, func(svc *v1.Service) {
+		svc.Spec.Template.Spec.Containers = containers
+	}); err != nil {
+		t.Fatalf("Failed to create initial Service %v: %v", names.Service, err)
+	}
+
+	// Validate State after Creation
+	if err := validateControlPlane(t, clients, names, "1" /*1 is the expected generation value*/); err != nil {
+		t.Error(err)
+	}
+
+	if err := validateDataPlane(t, clients, names, test.MultiContainerResponse); err != nil {
+		t.Error(err)
+	}
+}
